/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/* $Header: /a/helios/home/woodeaton/src/X11R5/mit/server/ddx/dec/pxpmax/RCS/pxpmax_io.c,v 1.4 1992/10/10 08:15:44 trq Exp trq $ */
/* $Log: pxpmax_io.c,v $
 * Revision 1.4  1992/10/10  08:15:44  trq
 * Use px_info->rb_phys instead of magic number.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/tty.h>
#include <errno.h>
#include <sys/devio.h>
#include <io/tc/gx.h>
#define N_MAX_CLIPRECTS       8
#define N_MAX_CLIPLISTS      16
#include <io/tc/ga.h>
#include <io/tc/stamp.h>
#include <io/tc/pmevent.h>
#include <io/tc/dc7085reg.h>

#include "misc.h"
#include "X.h"
#define NEED_EVENTS
#include "Xproto.h"
#include "scrnintstr.h"
#include "pixmap.h"
#include "input.h"
#include "cursorstr.h"
#include "regionstr.h"
#include "resource.h"
#include "dixstruct.h"

#include "pxpmaxcolor.h"
#include "mi.h"
#include "cfb.h"

static void pxpmaxQueryBestSize();
static Bool pxpmaxRealizeCursor();
static Bool pxpmaxUnrealizeCursor();
static Bool pxpmaxDisplayCursor();
static void pxpmaxRecolorCursor();
Bool pxpmaxSetCursorPosition();
static void pxpmaxCursorLimits();
static void pxpmaxConstrainCursor();
static void pxpmaxPointerNonInterestBox();
static void pxpmaxChangeKeyboardControl();
static void pxpmaxChangePointerControl();
static void pxpmaxClick();

extern void miRecolorCursor();
extern void NoopDDA();

/* XXX warning following symbol is copied from pmreg.h - this will
 * be fixed when the driver header files get straightened out. PK
 * 17 Aug 88
 */
#define MOTION_BUFFER_SIZE 100

/*
 * These "statics" imply there will never be more than one "pm" display and
 * mouse attached to a server; otherwise, they would be in SCREEN private info.
 *
 * This is probably the case since the Ultrix driver does not allow one to
 * open the display without also getting at the same file descriptor that
 * handles the mouse and the keyboard. Most of it is merely to pass information
 * to the keyboard and pointer devices. This could be done more cleanly by
 * repeating the ioctl to get the PM_Info.
 */

int			fdPM;
gxInfo		*px_info;

static ga_ComAreaPtr rb_cntrl;
static Com2dPtr rb_com2d;

static pmEventQueue	*queue;
static pmBox		*mbox;
static pmCursor		*mouse;
static int		qLimit;
static int		lastEventTime;
static DevicePtr	pmKeyboard;
static DevicePtr	pmPointer;
static int		hotX, hotY;
static BoxRec		cursorRange, cursorConstraint;
static int              dpix = -1, dpiy = -1, dpi = -1;
static char             *blackValue = NULL, *whiteValue = NULL;
static int		class = PseudoColor;
static unsigned	        cursColors[6]; /* 0-2 bg 3-5 fg */

#define MAX_LED 3	/* only 3 LED's can be set by user; Lock LED
			   is controlled by server */

#ifdef VSYNCFIXED
#define CURRENT_TIME	queue->timestamp_ms
#else
#define CURRENT_TIME	GetTimeInMillis()
#endif

#define NoSuchClass -1

static int
ParseClass(className)
    char *	className;
{
    static char *names[] = {
	"StaticGray", "GrayScale", "StaticColor",
	"PseudoColor", "TrueColor"};
    /* only the ones we support and must be in order from X.h, since
     * return value depends on index into array.
     */
    int i;
    for (i = 0; i < sizeof(names)/sizeof(char *); i++)
    {
	if (strcmp(names[i], className) == 0)
	    return i;
    }
    return NoSuchClass;
}

static Bool
commandLineMatch( argc, argv, pat, pmatch)
    int         argc;           /* may NOT be changed */
    char *      argv[];         /* may NOT be changed */
    char *      pat;
{
    int         ic;

    for ( ic=0; ic<argc; ic++)
        if ( strcmp( argv[ic], pat) == 0)
            return TRUE;
    return FALSE;
}

static Bool
commandLinePairMatch( argc, argv, pat, pmatch)
    int         argc;           /* may NOT be changed */
    char *      argv[];         /* may NOT be changed */
    char *      pat;
    char **     pmatch;         /* RETURN */
{
    int         ic;

    for ( ic=0; ic<argc; ic++)
        if ( strcmp( argv[ic], pat) == 0)
        {
            *pmatch = argv[ ic+1];
            return TRUE;
	}
    return FALSE;
}

static void
colorNameToColor( pname, pred, pgreen, pblue)
    char *      pname;
    u_int *     pred;
    u_int *     pgreen;
    u_int *     pblue;
{
    if ( *pname == '#')
    {
        pname++;                /* skip over # */
        sscanf( pname, "%2x", pred);
        *pred <<= 8;

        pname += 2;
        sscanf( pname, "%2x", pgreen);
        *pgreen <<= 8;

        pname += 2;
        sscanf( pname, "%2x", pblue);
        *pblue <<= 8;
    }
    else /* named color */
    {
        *pred = *pgreen = *pblue = 0; /*OsLookupColor thinks these are shorts*/
        OsLookupColor( 0 /*"screen", not used*/, pname, strlen( pname),
                pred, pgreen, pblue);
    }
}

/* SaveScreen does blanking, so no need to worry about the interval timer */

static Bool
pxpmaxSaveScreen(pScreen, on)
    ScreenPtr pScreen;
    int on;
{
    if (on == SCREEN_SAVER_FORCER)
    {
        lastEventTime = CURRENT_TIME;	
    }
    else if (on == SCREEN_SAVER_ON)
    {
	if (ioctl(fdPM, QIOVIDEOOFF) < 0)
	    ErrorF("pmSaveScreen: failed to turn screen off.\n");
    } else {
	if (ioctl(fdPM, QIOVIDEOON) < 0)
	    ErrorF("pmSaveScreen: failed to turn screen on.\n");
    }
    return TRUE;
}

Bool
pxpmaxScreenClose(index, pScreen)
    int index;
    ScreenPtr pScreen;
{
    /* This routine frees all of the dynamically allocated space associate
	with a screen. */

    /* ||| XXX cfbScreenClose(pScreen); */

/* ||| Just seeing if this lets me shut down and restart clients
    if(close(fdPM))
    {
	ErrorF("Close of fdPM yields %d\n", errno);
	return(FALSE);
    }
*/
    return(TRUE);
}

Bool
pxpmaxScreenInit(index, pScreen, argc, argv)
    int index;
    ScreenPtr pScreen;
    int argc;
    char **argv;
{
    register    PixmapPtr pPixmap;
    ColormapPtr pColormap;
    int		i;
    static int  mapOnce = FALSE;

/* for initializing color map entries */
    u_int blackred      = 0x0000;
    u_int blackgreen    = 0x0000;
    u_int blackblue     = 0x0000;

    u_int whitered      = 0xffff;
    u_int whitegreen    = 0xffff;
    u_int whiteblue     = 0xffff;

    if (mapOnce)
    {
	/*
	 * The reason you need to do this is so that when the
	 * the server recycles and the system is in video off
	 * state you need something to turn the video on.  Note
	 * that unlike the vaxstar you don't do a fresh open 
	 * and close of the device and thus the video remains
	 * off if you don't explicitly turn it on.
	 */
	if (ioctl(fdPM, QIOVIDEOON) < 0)
	    ErrorF("pmSaveScreen: failed to turn screen on.\n");
    }

    if (! mapOnce) {
	if ((fdPM = open("/dev/mouse", O_RDWR | O_NDELAY, 0)) < 0)
	{
	    ErrorF("couldn't open /dev/mouse \n");
	    return FALSE;
	}

	if (ioctl(fdPM, QIOCGINFO, &px_info) < 0)
	{
	    extern int errno;
	    int en = errno;
	    fprintf(stderr, "errno = %d, ", en);
	    ErrorF("error getting address of pmax screen\n");
	    close(fdPM);
	    return FALSE;
	}
	ioctl(fdPM, QIOKERNLOOP);

	mapOnce = TRUE;
    }

    rb_cntrl = (ga_ComAreaPtr)px_info->rb_addr;
    rb_com2d = &(rb_cntrl->SRV2DCom);
    
    queue = &px_info->qe;
    mouse = &px_info->mouse;
    qLimit = queue->eSize - 1;
    lastEventTime = CURRENT_TIME;

    /* discard all the current input events */
    queue->eHead = queue->eTail;

    /*
     * set keyclick, mouse acceleration and threshold
     */
    {
        int	    clicklevel;
        char *      clickvolume;
        char *      mouseAcceleration;
        int         ma = 4;
        char *      mouseThreshold;
        int         mt = 4;
        PtrCtrl     ctrl;

        if ( commandLinePairMatch( argc, argv, "c", &clickvolume))
	{
		sscanf( clickvolume, "%d", &clicklevel);
		pxpmaxClick(clicklevel);
	}

        if ( commandLineMatch( argc, argv, "-c"))
	{
		pxpmaxClick(0);
	}
	    
        /*
         * calling pxpmaxChangePointerControl here may be unclean       XXX
         */
	if ( commandLinePairMatch( argc, argv, "-a", &mouseAcceleration))
	    sscanf( mouseAcceleration, "%d", &ma);
	if ( commandLinePairMatch( argc, argv, "-t", &mouseThreshold))
	    sscanf( mouseThreshold, "%d", &mt);
	ctrl.num = ma;
	ctrl.den = 1;
	ctrl.threshold = mt;
	pxpmaxChangePointerControl( (DevicePtr) NULL, &ctrl);
    }


    cursorRange.x1 = -15;
    cursorRange.x2 = px_info->max_x - 1;
    cursorRange.y1 = -15;
    cursorRange.y2 = px_info->max_y - 1;

/* XXX FRAMEWIDTH is bogus.  Don't we get this with initial IOCTL?? */

    if (dpi == -1) /* dpi has not been set */
    {
        if (dpix == -1) /* ie dpix has not been set */
        {
	    if (dpiy == -1)
	    {
	        dpix = 100;
	        dpiy = 100;
	    }
	    else
	        dpix = dpiy;
        }
        else
        { 
	    if (dpiy == -1)
	        dpiy = dpix;
        }
    }
    else
    {
	dpix = dpi;
	dpiy = dpi;
    }

    pScreen->CloseScreen = pxpmaxScreenClose;
    pScreen->SaveScreen = pxpmaxSaveScreen;
    pScreen->RealizeCursor = pxpmaxRealizeCursor;
    pScreen->UnrealizeCursor = pxpmaxUnrealizeCursor;
    pScreen->DisplayCursor = pxpmaxDisplayCursor;
    pScreen->SetCursorPosition = pxpmaxSetCursorPosition;
    pScreen->CursorLimits = pxpmaxCursorLimits;
    pScreen->PointerNonInterestBox = pxpmaxPointerNonInterestBox;
    pScreen->ConstrainCursor = pxpmaxConstrainCursor;
    pScreen->RecolorCursor = pxpmaxRecolorCursor;
    pScreen->QueryBestSize = pxpmaxQueryBestSize;

    pScreen->StoreColors = pxpmaxStoreColors;
    pScreen->InstallColormap = pxpmaxInstallColormap;
    pScreen->UninstallColormap = pxpmaxUninstallColormap;
    pScreen->ListInstalledColormaps = pxpmaxListInstalledColormaps;

/* XXX this is the hard part */
    if (!cfbScreenInit(pScreen, NULL, px_info->max_x, px_info->max_y, dpix,
		       dpiy, px_info->max_x, class, px_info->max_x)
	|| !cfbCreateDefColormap (pScreen))
    {
	close (fdPM);
	return FALSE;
    }
    return TRUE;
}

static void
ChangeLED(led, on)
    int led;
    Bool on ;
{
    struct gx_kpcmd ioc;

    switch (led) {
       case 1:
          ioc.par[0] = LED_1;
          break;
       case 2:
          ioc.par[0] = LED_2;
          break;
       case 3:
          /* the keyboard's LED_3 is the Lock LED, which the server owns.
             So the user's LED #3 maps to the keyboard's LED_4. */
          ioc.par[0] = LED_4;
          break;
       default:
          return;   /* out-of-range LED value */
          }

    ioc.cmd = on ? LK_LED_ENABLE : LK_LED_DISABLE;
    ioc.par[1] = 0;
    ioc.nbytes = 1;
    ioctl(fdPM, QIOCKPCMD, &ioc);
}


static void
pxpmaxClick(click)
    int         click;
{
#define LK_ENABLE_CLICK 0x1b	/* enable keyclick / set volume	*/
#define LK_DISABLE_CLICK 0x99	/* disable keyclick entirely	*/
#define LK_ENABLE_BELL 0x23	/* enable bell / set volume 	*/

    struct gx_kpcmd ioc;

    if (click == 0)    /* turn click off */
    {
	ioc.nbytes = 0;
	ioc.cmd = LK_DISABLE_CLICK;
	ioctl(fdPM, QIOCKPCMD, &ioc);
    }
    else 
    {
        int volume;

        volume = 7 - ((click / 14) & 7);
	ioc.nbytes = 1;
	ioc.cmd = LK_ENABLE_CLICK;
	ioc.par[0] = volume;
	ioctl(fdPM, QIOCKPCMD, &ioc);
    }
#undef LK_ENABLE_CLICK 
#undef LK_DISABLE_CLICK
#undef LK_ENABLE_BELL 

}

static void
pxpmaxChangeKeyboardControl(device, ctrl)
    DevicePtr device;
    KeybdCtrl *ctrl;
{
    int i;

    pxpmaxClick(ctrl->click);

    /* ctrl->bell: the DIX layer handles the base volume for the bell */
    
    /* ctrl->bell_pitch: as far as I can tell, you can't set this on lk201 */

    /* ctrl->bell_duration: as far as I can tell, you can't set this  */

    /* LEDs */
    for (i=1; i<=MAX_LED; i++)
        ChangeLED(i, (ctrl->leds & (1 << (i-1))));

    /* ctrl->autoRepeat: I'm turning it all on or all off.  */

    SetLKAutoRepeat(ctrl->autoRepeat);
}

static void
pxpmaxBell(loud, pDevice)
    int loud;
    DevicePtr pDevice;
{
    struct gx_kpcmd ioc;

/* the lk201 volume is between 7 (quiet but audible) and 0 (loud) */
    loud = 7 - ((loud / 14) & 7);
    ioc.nbytes = 1;
    ioc.cmd = LK_BELL_ENABLE;
    ioc.par[0] = loud;
    ioctl(fdPM, QIOCKPCMD, &ioc);

    ioc.nbytes = 0;
    ioc.cmd = LK_RING_BELL;
    ioctl(fdPM, QIOCKPCMD, &ioc);
}

/*
 * These serve protocol requests, setting/getting acceleration and threshold.
 * X10 analog is "SetMouseCharacteristics".
 */
static void
pxpmaxChangePointerControl(device, ctrl)
    DevicePtr device;
    PtrCtrl   *ctrl;
{
    px_info->mthreshold = ctrl->threshold;
    if (!(px_info->mscale = ctrl->num / ctrl->den))
	px_info->mscale = 1;	/* watch for den > num */
}

static int
pxpmaxGetMotionEvents(pDevice, buff, start, stop)
    CARD32 start, stop;
    DevicePtr pDevice;
    xTimecoord *buff;
{
    int     count = 0;
    int     tcFirst = queue->tcNext;
    int     tcLast = (tcFirst) ? tcFirst - 1 : queue->tcSize - 1;
    register    pmTimeCoord * tcs;
    int     i;

    for (i = tcFirst;; i++) {
	if (i = queue->tcSize)
	    i = 0;
	tcs = &queue->tcs[i];
	if ((start <= tcs->time) && (tcs->time <= stop)) {
	    buff[count].time = tcs->time;
	    buff[count].x = tcs->x;
	    buff[count].y = tcs->y;
	    count++;
	}
	if (i == tcLast)
	    return count;
    }
}

int
pxpmaxMouseProc(pDev, onoff, argc, argv)
    DevicePtr pDev;
    int onoff, argc;
    char *argv[];
{
    int     i;
    BYTE    map[4];

    switch (onoff)
    {
	case DEVICE_INIT: 
	    pmPointer = pDev;
	    pDev->devicePrivate = (pointer) &queue;
	    map[1] = 1;
	    map[2] = 2;
	    map[3] = 3;
	    InitPointerDeviceStruct(
		pmPointer, map, 3, pxpmaxGetMotionEvents,
		pxpmaxChangePointerControl, MOTION_BUFFER_SIZE);
	    SetInputCheck(&queue->eHead, &queue->eTail);
	    hotX = hotY = 0;	
	    break;
	case DEVICE_ON: 
	    pDev->on = TRUE;
	    AddEnabledDevice(fdPM); 
	    break;
	case DEVICE_OFF: 
	    pDev->on = FALSE;
	    RemoveEnabledDevice(fdPM);
	    break;
	case DEVICE_CLOSE:
	    break;
    }
    return Success;
}

#define LK_REPEAT_ON  0xe3
#define LK_REPEAT_OFF 0xe1

int
SetLKAutoRepeat (onoff)
    Bool onoff;
{
    extern char *AutoRepeatLKMode();
    extern char *UpDownLKMode();
    
    struct gx_kpcmd ioc;
    register char  *divsets;
    divsets = onoff ? (char *) AutoRepeatLKMode() : (char *) UpDownLKMode();
    ioc.nbytes = 0;
    while (ioc.cmd = *divsets++)
	ioctl(fdPM, QIOCKPCMD, &ioc);
    ioc.cmd = ((onoff > 0) ? LK_REPEAT_ON : LK_REPEAT_OFF);
    return(ioctl(fdPM, QIOCKPCMD, &ioc));
}

int
pxpmaxKeybdProc(pDev, onoff, argc, argv)
    DevicePtr pDev;
    int onoff, argc;
    char *argv[];
{
    KeySymsRec keySyms;
    CARD8 modMap[MAP_LENGTH];

    switch (onoff)
    {
	case DEVICE_INIT: 
/* Note that keyclick is off by default.  The QDSS MIT server sets it
   to 20 */	
	    pmKeyboard = pDev;
	    GetLK201Mappings( &keySyms, modMap);
	    InitKeyboardDeviceStruct(
		    pmKeyboard, &keySyms, modMap, pxpmaxBell,
		    pxpmaxChangeKeyboardControl);
            pxpmaxClick(0);     /* turn key click off */	    
	    /* Free the key sym mapping space allocated by GetLK201Mappings. */
	    Xfree(keySyms.map);
	    break;
	case DEVICE_ON: 
	    pDev->on = TRUE;
	    AddEnabledDevice(fdPM); 
	    break;
	case DEVICE_OFF: 
	    pDev->on = FALSE;
	    RemoveEnabledDevice(fdPM);
	    break;
	case DEVICE_CLOSE: 
	    break;
    }
    return Success;
}
/*
 * The driver has been set up to put events in the queue that are identical
 * in shape to the events that the DDX layer has to deliver to ProcessInput
 * in DIX.
 */
extern int screenIsSaved;

void
ProcessInputEvents()
{
#define DEVICE_KEYBOARD 2
    xEvent x;
    pmEvent e;
    register int    i;

    i = queue->eHead;
    while (i != queue->eTail)
    {
	e = *((pmEvent *)&queue->events[i]);
	if (screenIsSaved == SCREEN_SAVER_ON)
	    SaveScreens(SCREEN_SAVER_OFF, ScreenSaverReset);
	x.u.keyButtonPointer.rootX = e.x + hotX;
	x.u.keyButtonPointer.rootY = e.y + hotY;
	x.u.keyButtonPointer.time = lastEventTime = e.time;
	x.u.u.detail = e.key;

	if (e.device == DEVICE_KEYBOARD)
	{
	    switch (e.type)
	    {
		case BUTTON_DOWN_TYPE: 
		    x.u.u.type = KeyPress;
		    (*pmKeyboard->processInputProc) (&x, pmKeyboard, 1);
		    break;
		case BUTTON_UP_TYPE: 
		    x.u.u.type = KeyRelease;
		    (*pmKeyboard->processInputProc) (&x, pmKeyboard, 1);
		    break;
		default: 	       /* hopefully BUTTON_RAW_TYPE */
		    ProcessLK201Input(&x, pmKeyboard);
	    }
	}
	else
	{
	    switch (e.type)
	    {
		case BUTTON_DOWN_TYPE: 
		    x.u.u.type = ButtonPress;
		    break;
		case BUTTON_UP_TYPE: 
		    x.u.u.type = ButtonRelease;
		    break;
		case MOTION_TYPE: 
		    x.u.u.type = MotionNotify;
		    break;
		default: 
		    printf("Unknown input event = %d\n",e.type);
		    continue;
	    }
	    (*pmPointer->processInputProc) (&x, pmPointer, 1);
	}
	if (i == qLimit)
	    i = queue->eHead = 0;
	else
	    i = ++queue->eHead;
    }
#undef DEVICE_KEYBOARD
}

TimeSinceLastInputEvent()
{
    if (lastEventTime == 0)
	lastEventTime = CURRENT_TIME;
    return CURRENT_TIME -  lastEventTime;
}

/*
 * set the bounds in the device for this particular cursor
 */
static void
pxpmaxConstrainCursor( pScr, pBox)
    ScreenPtr	pScr;
    BoxPtr	pBox;
{
    cursorConstraint = *pBox;
    if (px_info)
    {
	px_info->max_cur_x = pBox->x2 - hotX - 1;
	px_info->max_cur_y = pBox->y2 - hotY - 1;
	px_info->min_cur_x = pBox->x1 - hotX;
	px_info->min_cur_y = pBox->y1 - hotY;
    }
}

Bool
pxpmaxSetCursorPosition( pScr, newx, newy, generateEvent)
    ScreenPtr	pScr;
    unsigned int	newx;
    unsigned int	newy;
    Bool		generateEvent;
{
    pmCursor pmCPos;
    xEvent motion;

    pmCPos.x = newx - hotX;
    pmCPos.y = newy - hotY;

    if (ioctl (fdPM, QIOCGXSTATE, &pmCPos) < 0) {
	ErrorF ("error warping cursor\n");
	return FALSE;
    }

    if (generateEvent) {
	if (queue->eHead != queue->eTail)
	    ProcessInputEvents ();
	motion.u.keyButtonPointer.rootX = newx;
	motion.u.keyButtonPointer.rootY = newy;
	motion.u.keyButtonPointer.time = currentTime.milliseconds;
	motion.u.u.type = MotionNotify;
	(pmPointer->processInputProc)(&motion, pmPointer, 1);
    }
    return TRUE;
}

static Bool
pxpmaxDisplayCursor( pScr, pCurs)
    ScreenPtr	pScr;
    CursorPtr	pCurs;
{
    register    int		i;
    int		x, y;

    /*
     * load the cursor
     */
    if ((hotX != pCurs->bits->xhot) || (hotY != pCurs->bits->yhot))
    {
	x = mouse->x + hotX;
	y = mouse->y + hotY;
	hotX = pCurs->bits->xhot;
	hotY = pCurs->bits->yhot;
	pxpmaxSetCursorPosition(pScr, x, y, FALSE);
	pxpmaxConstrainCursor(pScr, &cursorConstraint);
		/* to update constraints in driver */
    }
    bcopy(pCurs->devPriv[pScr->myNum], px_info->curs_bits,
	  GX_CURSORBYTES);
    
    if ( ioctl( fdPM, QIO_WCURSOR) < 0)
    {
	ErrorF( "error loading bits of new cursor\n");
        return FALSE;
    }

    /* pick the background color */
    px_info->curs_bg = ((pCurs->backRed & 0xff00) << 8)
      | (pCurs->backGreen & 0xff00) | ((pCurs->backBlue & 0xff00) >> 8);
    
    /* pick the foreground color */
    px_info->curs_fg = ((pCurs->foreRed & 0xff00) << 8)
      | (pCurs->foreGreen & 0xff00) | ((pCurs->foreBlue & 0xff00) >> 8);

    if ( ioctl( fdPM, QIO_WCCOLOR, cursColors) < 0)
    {
	ErrorF( "error writing colors of new cursor\n");
        return FALSE;
    }

    return (TRUE);

}

static void
pxpmaxRecolorCursor (pScr, pCurs, displayed)
    ScreenPtr	pScr;
    CursorPtr	pCurs;
    Bool	displayed;
{
    if (!displayed)
	return;

    /* pick the background color */
    px_info->curs_bg = ((pCurs->backRed & 0xff00) << 8)
      | (pCurs->backGreen & 0xff00) | ((pCurs->backBlue & 0xff00) >> 8);
    
    /* pick the foreground color */
    px_info->curs_fg = ((pCurs->foreRed & 0xff00) << 8)
      | (pCurs->foreGreen & 0xff00) | ((pCurs->foreBlue & 0xff00) >> 8);

    if ( ioctl( fdPM, QIO_WCCOLOR) < 0)
    {
	ErrorF( "error writing colors of new cursor\n");
        return;
    }

    return;
}

static Bool
pxpmaxRealizeCursor( pScr, pCurs)
    ScreenPtr	pScr;
    CursorPtr	pCurs;	/* a SERVER-DEPENDENT cursor */
{
    int	forecolor = 0;
    int	backcolor = ~forecolor;
    register char	*a;	/* vaxstar-defined */
    register char *	mask;	/* server-defined */
    register char *	src;	/* server-defined */
    int		i, j, k;
    int		lastRow
      = ((pCurs->bits->height < GX_CURH) ? pCurs->bits->height : GX_CURH);
    int w_b = (pCurs->bits->width+7)/8;

    pCurs->devPriv[ pScr->myNum] = (pointer)Xalloc(GX_CURSORBYTES);
    bzero((char *)pCurs->devPriv[ pScr->myNum], GX_CURSORBYTES);

    /*
     * munge the SERVER-DEPENDENT, device-independent cursor bits into
     * what the device wants, which is 32 contiguous shorts.
     *
     * cursor hardware has "A" and "B" bitmaps
     * logic table is:
     *
     *		A	B	cursor
     *
     *		0	0	transparent
     *		1	0	xor (not used)
     *		0	1	black
     *		1	1	white
     */

    /*
     * "a" bitmap = image 
     */
    /*
     * "b" bitmap can be same as "mask", providing "a" is never on when
     *  "b" is off.
     */
    for (i=0; i < lastRow; i++)
    {
				/* src and mask are int padded */
      src = (char *)pCurs->bits->source + i*(w_b/4+1)*4;
      mask = (char *)pCurs->bits->mask + i*(w_b/4+1)*4;
      a = ((char *)pCurs->devPriv[pScr->myNum]) + i*GX_CURW/4;
      for(j = 0; j < w_b; j++, src++, mask++)
	{
	  for(k = 0; k < 4; k++)
	    {
	      *a |= ((*mask & (1 << k)) << (7-k)) >> (2*k);
	      *a |= ((*src & (1 << k)) << (6-k)) >> (2*k);
	    }
	  a++;
	  for(k = 0; k < 4; k++)
	    {
	      *a |= ((*mask & (1 << (k+4))) << (3-k)) >> (2*k);
	      *a |= ((*src & (1 << (k+4))) << (3-k)) >> (1 + 2*k);
	    }
	  a++;
	}
    }
    return TRUE;
}

static Bool
pxpmaxUnrealizeCursor( pScr, pCurs)
    ScreenPtr	pScr;
    CursorPtr	pCurs;
{
    Xfree( pCurs->devPriv[ pScr->myNum]);
    return TRUE;
}

/*
 * pm cursor top-left corner can now go to negative coordinates
 */
static void
pxpmaxCursorLimits( pScr, pCurs, pHotBox, pPhysBox)
    ScreenPtr	pScr;
    CursorPtr	pCurs;
    BoxPtr	pHotBox;
    BoxPtr	pPhysBox;	/* return value */
{
    pPhysBox->x1 = max( pHotBox->x1, cursorRange.x1 + (int) pCurs->bits->xhot);
    pPhysBox->y1 = max( pHotBox->y1, cursorRange.y1 + (int) pCurs->bits->yhot);
    pPhysBox->x2 = min( pHotBox->x2, cursorRange.x2 + 1);
    pPhysBox->y2 = min( pHotBox->y2, cursorRange.y2 + 1);
}

static void
pxpmaxPointerNonInterestBox( pScr, pBox)
    ScreenPtr	pScr;
    BoxPtr	pBox;
{
    px_info->mbox.bottom = pBox->y2;
    px_info->mbox.top = pBox->y1;
    px_info->mbox.left = pBox->x1;
    px_info->mbox.right = pBox->x2;
}

static void
pxpmaxQueryBestSize(class, pwidth, pheight)
    int class;
    short *pwidth;
    short *pheight;
{
    unsigned width, test;

    if (*pwidth > 0)
    {
      switch(class)
      {
        case CursorShape:
	  *pwidth = 16;
	  *pheight = 16;
	  break;
        case TileShape:
        case StippleShape:
	  width = *pwidth;
	  /* Return the closes power of two not less than what they gave me */
	  test = 0x80000000;
	  /* Find the highest 1 bit in the width given */
	  while(!(test & width))
	     test >>= 1;
	  /* If their number is greater than that, bump up to the next
	   *  power of two */
	  if((test - 1) & width)
	     test <<= 1;
	  *pwidth = test;
	  /* We don't care what height they use */
	  break;
       }
    }
}

void
SetLockLED (on)
    Bool on;
{
    struct gx_kpcmd ioc;
    ioc.cmd = on ? LK_LED_ENABLE : LK_LED_DISABLE;
    ioc.par[0] = LED_3;
    ioc.par[1] = 0;
    ioc.nbytes = 1;
    ioctl(fdPM, QIOCKPCMD, &ioc);
}

/*
 * DDX - specific abort routine.  Called by AbortServer().
 */
void
AbortDDX()
{
}

/* Called by GiveUp(). */
void
ddxGiveUp()
{
}

int
ddxProcessArgument (argc, argv, i)
    int argc;
    char *argv[];
    int i;
{
    int			argind=i;
    int			skip;
    static int		Once=0;
    void		ddxUseMsg();

    skip = 0;
    if (!Once)
    {
        blackValue = NULL;
        whiteValue = NULL;
	Once = 1;
    }

    if (strcmp( argv[argind], "-dpix") == 0)
    {
	if (++argind < argc)
	{
	    dpix = atoi(argv[argind]);
	    skip = 2;
	}
	else
	    return 0;	/* failed to parse */
    }
    else if (strcmp( argv[argind], "-dpiy") == 0)
    {
	if (++argind < argc)
	{
	    dpiy = atoi(argv[argind]);
	    skip = 2;
	}
	else
	    return 0;
    }
    else if (strcmp( argv[argind], "-dpi") == 0)
    {
	if (++argind < argc)
	{
	    dpi = atoi(argv[argind]);
	    dpix = dpi;
	    dpiy = dpi;
	    skip = 2;
	}
	else
	    return 0;
    }
    else 
    if(strcmp(argv[argind], "-bp") == 0 )
    {
	    if(++argind < argc)
	    {
		blackValue = argv[argind];
		skip = 2;
	    }
	    else
	        return 0;
    }
    else
    if(strcmp(argv[argind], "-wp") == 0 )
    {
	    if(++argind < argc)
	    {
		whiteValue = argv[argind];
		skip = 2;
	    }
	    else
	        return 0;
    }
    else
    if (strcmp(argv[argind], "-class") == 0 )
    {
	    if(++argind < argc)
	    {
		class = ParseClass(argv[argind]);
		if (class == NoSuchClass)
		    return 0;
		skip = 2;
	    }
	    else
	        return 0;
    }
    return skip;
}

void
ddxUseMsg()
{
    ErrorF ("\n");
    ErrorF ("\n");
    ErrorF ("Device Dependent Usage\n");
    ErrorF ("\n");
    ErrorF ("-dpix <n>          Dots per inch, x coordinate\n");
    ErrorF ("-dpiy <n>          Dots per inch, y coordinate\n");
    ErrorF ("-dpi <n>           Dots per inch, x and y coordinates\n");
    ErrorF ("                   (overrides -dpix and -dpiy above)\n");
    ErrorF ("-bp #XXX           color for BlackPixel for screen\n");
    ErrorF ("-wp #XXX           color for WhitePixel for screen\n");
    ErrorF ("-class <classname> type of Visual for root window\n");
    ErrorF ("       one of StaticGray, StaticColor, PseudoColor,\n");
    ErrorF ("       GrayScale, or even TrueColor!\n");
}

ga_send_packet()
{
  int nbuf;
  int gaintr;
  int i;
  
  for(i = 0; rb_com2d->intr_status != 0; i++);
  
/*  if(i != 0)
    fprintf(stderr, "Spun on intr_status: %d\n", i);
*/

  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return FALSE;
    }
  if(nbuf != GA_QUEUE_PACKETS-1)
      rb_com2d->lastWritten++;
  else
      rb_com2d->lastWritten = 0;
  
  if(rb_com2d->lastRead == 0)
      rb_com2d->lastRead = GA_QUEUE_PACKETS-1;
  else
      rb_com2d->lastRead--;

  rb_com2d->intr_status |= 1;
	
  gaintr = GAO_STIC(px_info)->ipdvint;
	
  gaintr |= (STIC_INT_P_EN| STIC_INT_P | STIC_INT_P_WE);
  gaintr &= ~(STIC_INT_E_WE|STIC_INT_V_WE);
	
  GAO_STIC(px_info)->ipdvint = gaintr;
}

#ifdef __GNUC__
__inline__
#endif
static int aluxtopx(alu)
     int alu;
{
  int retval;
  switch(alu)
    {
    case GXclear:
      retval = UMET_CLEAR;
      break;
    case GXand:
      retval = UMET_AND;
      break;
    case GXandReverse:
      retval = UMET_ANDREVERSE;
      break;
    case GXcopy:
      retval = UMET_COPY;
      break;
    case GXandInverted:
      retval = UMET_ANDINVERTED;
      break;
    case GXnoop:
      retval = UMET_NOOP;
      break;
    case GXxor:
      retval = UMET_XOR;
      break;
    case GXor:
      retval = UMET_OR;
      break;
    case GXnor:
      retval = UMET_NOR;
      break;
    case GXequiv:
      retval = UMET_EQUIV;
      break;
    case GXinvert:
      retval = UMET_INVERT;
      break;
    case GXorReverse:
      retval = UMET_ORREVERSE;
      break;
    case GXcopyInverted:
      retval = UMET_COPYINVERTED;
      break;
    case GXorInverted:
      retval = UMET_ORINVERTED;
      break;
    case GXnand:
      retval = UMET_NAND;
      break;
    case GXset:
      retval = UMET_SET;
      break;
    default:
      fprintf(stderr, "unhandled alu\n");
      retval = UMET_COPY;
    }
  return retval;
}
  
pxpmax1Line(x1, y1, x2, y2, pixel, alu, planemask)
     int x1, y1, x2, y2;
     unsigned long pixel;
     int alu;
     int planemask;
{
  int *nbuf_addr;
  int nbuf;

  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return FALSE;
    }
  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  nbuf_addr[0] = 5;
  nbuf_addr[1] = 10;
  nbuf_addr[2] = 16;
  nbuf_addr[3] = STAMP_CMD_LINES | RGB_CONST | LW_PERPKT | CLIPRECT;
  planemask &= 0xff;
  nbuf_addr[4] = (1 << 24) | (planemask << 16) | (planemask << 8) | planemask;
  nbuf_addr[5] = 0;
  nbuf_addr[6] = aluxtopx(alu);
  nbuf_addr[6] |= UPD_ENABLE | WE_CLIPRECT;
  nbuf_addr[7] = 4;
  nbuf_addr[8] = 0x7c00400;
  nbuf_addr[9] = 0x7ffffff;
  nbuf_addr[10] = pixel;
  nbuf_addr[11] = (x1 << 19) | (y1 << 3);
  nbuf_addr[12] = (x2 << 19) | (y2 << 3);
  
  ga_send_packet();
}  

static int line_count;
static int *line_addr;

pxpmaxStartLines(pixel, alu, planemask)
     unsigned long pixel;
     int alu;
     int planemask;
{
  int nbuf;

  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return FALSE;
    }
  line_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  line_addr[0] = 5;
  line_addr[2] = 16;
  line_addr[3] = STAMP_CMD_LINES | RGB_CONST | LW_PERPKT | CLIPRECT;
  planemask &= 0xff;
  line_addr[4] = (planemask << 16) | (planemask << 8) | planemask;
  line_addr[5] = 0;
  line_addr[6] = aluxtopx(alu);
  line_addr[6] |= UPD_ENABLE | WE_CLIPRECT;
  line_addr[7] = 4;
  line_addr[8] = 0x7c00400;
  line_addr[9] = 0x7ffffff;
  line_addr[10] = pixel;
  line_count = 0;
}  

pxpmaxAddLine(x1, y1, x2, y2)
     int x1, y1, x2, y2;
{
  line_addr[11+line_count*2] = (x1 << 19) | (y1 << 3);
  line_addr[12+line_count*2] = (x2 << 19) | (y2 << 3);
  line_count++;
  
  if(line_count >= STAMP_MAX_CMDS)
    {
      int planemask, pixel, alu;
      int nbuf;
      
      planemask = line_addr[4];
      alu = line_addr[6];
      pixel = line_addr[10];
      line_addr[1] = 8 + line_count*2;
      line_addr[4] |= (line_count << 24);
      
      ga_send_packet();

      nbuf = rb_com2d->lastWritten;
      if(nbuf >= GA_QUEUE_PACKETS)
	{
	  ErrorF("Too many packets\n");
	  return FALSE;
	}
      line_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
      line_addr[0] = 5;
      line_addr[2] = 16;
      line_addr[3] = STAMP_CMD_LINES | RGB_CONST | LW_PERPKT | CLIPRECT;
      line_addr[4] = planemask;
      line_addr[5] = 0;
      line_addr[6] = alu;
      line_addr[7] = 4;
      line_addr[8] = 0x7c00400;
      line_addr[9] = 0x7ffffff;
      line_addr[10] = pixel;
      line_count = 0;
    }  
}  

pxpmaxFinishLines()
{
  if(line_count > 0)
    {
      line_addr[1] = 8 + line_count*2;
      line_addr[4] |= (line_count << 24);
      
      ga_send_packet();
    }
}

pxpmaxWideLine(pBox, pixel, alu, planemask)
     BoxPtr pBox;
     unsigned long pixel;
     int alu;
     int planemask;
{
  int *nbuf_addr;
  int nbuf;
  int y2m;

  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return FALSE;
    }
  
  y2m = pBox->y2 - 1;
  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  nbuf_addr[0] = 5;
  nbuf_addr[1] = 8;
  nbuf_addr[2] = 16;
  nbuf_addr[3] = STAMP_CMD_LINES | RGB_CONST | LW_PERPRIM;
  planemask &= 0xff;
  nbuf_addr[4] = (1 << 24) | (planemask << 16) | (planemask << 8) | planemask;
  nbuf_addr[5] = 0;
  nbuf_addr[6] = aluxtopx(alu);
  nbuf_addr[6] |= UPD_ENABLE;
  nbuf_addr[7] = pixel;
  nbuf_addr[8] = ((int)pBox->x1) << 19 | ((y2m + pBox->y1) << 2) | 0x3;
  nbuf_addr[9] = ((pBox->x2-1) << 19) | 0x70000
                 | ((y2m + pBox->y1) << 2) | 0x3;
  nbuf_addr[10] = ((y2m - pBox->y1) << 2) | 0x3;
  
  ga_send_packet();
}

static int bufno = 1;
#define GA_SPBUFSIZE (PAGE + HALF_PAGE)

void
pxpmaxPutSpan(y, xOrigin, xStart, xEnd, psrc, alu, planemask)
     int y;
     int xOrigin;
     int xStart;
     int xEnd;
     int *psrc;
     int alu;
     int planemask;
{
  int *nbuf_addr;
  int nbuf;
  int i;
  int image_off, span_len;

  span_len = xEnd - xStart;
  if((span_len >= GA_SPBUFSIZE) || (span_len < 0))
    {
      ErrorF("Span too large or negative\n");
      return;
    }
  if((xStart < 0) || (y < 0) || (xEnd > px_info->max_x) || (y >=
     px_info->max_y))
    {
      ErrorF("PutSpan: bad region\n");
      return;
    }
  
  bufno ^= 1;
  
  for(i = 0; i < span_len; i++)
    {
      rb_cntrl->image_buf[((bufno*GA_SPBUFSIZE)>>2) + i]
	= ((char *)psrc)[xOrigin+i];
    }

  image_off = bufno*GA_SPBUFSIZE + GA_IMAGE_BUFFER_OFF - 12;

  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return;
    }
  
  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  nbuf_addr[0] = 5;
  nbuf_addr[1] = 8;
  nbuf_addr[2] = 16;
  nbuf_addr[3] = STAMP_CMD_WRITESPANS;
  planemask &= 0xff;
  nbuf_addr[4] = (1 << 24) | (planemask << 16) | (planemask << 8) | planemask;
  nbuf_addr[5] = 0;
  nbuf_addr[6] = aluxtopx(alu);
  nbuf_addr[6] |= UPD_ENABLE | SPAN_MASK;
  nbuf_addr[7] = px_info->rb_phys | image_off;
  nbuf_addr[8] = 2*span_len*sizeof(int);
  nbuf_addr[9] = xStart << 19 | (y << 3);
  nbuf_addr[10] = 0x0;
  
  ga_send_packet();
}

void
pxpmaxPutBox(yStart, yEnd, xOrigin, xStart, xEnd, psrc, psrcWidth,
	     alu, planemask)
     int yStart;
     int yEnd;
     int xOrigin;
     int xStart;
     int xEnd;
     int *psrc;
     int psrcWidth;
     int alu;
     int planemask;
{
  int *nbuf_addr;
  int nbuf;
  int i;
  int image_off, span_len;
  int y;
  int ncom, span_off, tot_span_len;

  span_len = xEnd - xStart;
  if((span_len >= GA_SPBUFSIZE) || (span_len < 0))
    {
      ErrorF("Span too large or negative\n");
      return;
    }
  if((xStart < 0) || (yStart < 0) || (xEnd > px_info->max_x)
     || (yEnd > px_info->max_y))
    {
      ErrorF("PutSpan: bad region\n");
      return;
    }
  
  bufno ^= 1;
  
  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return;
    }
  ncom = 0;
  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  span_off = (bufno*GA_SPBUFSIZE)>>2;
  tot_span_len = 0;
  image_off = bufno*GA_SPBUFSIZE + GA_IMAGE_BUFFER_OFF - 12;
  for(y = yStart; y < yEnd; y++, psrc += psrcWidth)
    {
      for(i = 0; i < span_len; i++)
	rb_cntrl->image_buf[span_off+i] = ((char *)psrc)[xOrigin+i];
      nbuf_addr[7+ncom*4] = px_info->rb_phys | (image_off + ncom*4*span_len);
      nbuf_addr[8+ncom*4] = 2*span_len*sizeof(int);
      nbuf_addr[9+ncom*4] = xStart << 19 | (y << 3);
      nbuf_addr[10+ncom*4] = 0x0;
      ncom++;
      span_off += span_len;
      tot_span_len += span_len;
      if(y == yEnd - 1 || ncom >= STAMP_MAX_CMDS-1
	 || tot_span_len + span_len >= GA_SPBUFSIZE>>2)
	{
	  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
	  nbuf_addr[0] = 5;
	  nbuf_addr[1] = 4 + ncom*4;
	  nbuf_addr[2] = 16;
	  nbuf_addr[3] = STAMP_CMD_WRITESPANS;
	  planemask &= 0xff;
	  nbuf_addr[4] = (ncom << 24) | (planemask << 16)
	    | (planemask << 8) | planemask;
	  nbuf_addr[5] = 0;
	  nbuf_addr[6] = aluxtopx(alu);
	  nbuf_addr[6] |= UPD_ENABLE | SPAN_MASK;
	  ga_send_packet();

	  if(y != yEnd - 1)
	    {
	      bufno ^= 1;
	      nbuf = rb_com2d->lastWritten;
	      if(nbuf >= GA_QUEUE_PACKETS)
		{
		  ErrorF("Too many packets\n");
		  return;
		}
	      ncom = 0;
	      nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
	      span_off = (bufno*GA_SPBUFSIZE)>>2;
	      tot_span_len = 0;
	      image_off = bufno*GA_SPBUFSIZE + GA_IMAGE_BUFFER_OFF - 12;
	    }
	}
    }
}

void
pxpmaxGetSpan(y, xOrigin, xStart, xEnd, psrc, planemask)
     int y;
     int xOrigin;
     int xStart;
     int xEnd;
     int *psrc;
     int planemask;
{
  int *nbuf_addr;
  int nbuf;
  int i;
  int image_off, span_len;

  span_len = xEnd - xStart;
  if((span_len >= GA_SPBUFSIZE) || (span_len < 0))
    {
      ErrorF("Span too large or negative\n");
      return;
    }
  if((xStart < 0) || (y < 0) || (xEnd > px_info->max_x) || (y >=
     px_info->max_y))
    {
      ErrorF("GetSpan: bad region\n");
      return;
    }
  
  bufno ^= 1;
  
  image_off = bufno*GA_SPBUFSIZE + GA_IMAGE_BUFFER_OFF - 12;

  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return;
    }
  
  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  nbuf_addr[0] = 5;
  nbuf_addr[1] = 8;
  nbuf_addr[2] = 16;
  nbuf_addr[3] = STAMP_CMD_READSPANS;
/* is this right? */
  planemask = 0xff;
  nbuf_addr[4] = (1 << 24) | (planemask << 16) | (planemask << 8) | planemask;
  nbuf_addr[5] = 0;
  nbuf_addr[6] = UMET_NOOP | UPD_ENABLE | SPAN_MASK;
  nbuf_addr[7] = 0x40000000 | px_info->rb_phys | image_off;
  nbuf_addr[8] = 2*span_len*sizeof(int);
  nbuf_addr[9] = xStart << 19 | (y << 3);
  nbuf_addr[10] = 0x0;
  
  ga_send_packet();

/* wait for command to complete */
  for(i = 0; rb_com2d->intr_status != 0; i++);

  for(i = 0; i < span_len; i++)
    {
      ((unsigned char *)psrc)[xOrigin+i]
	= (rb_cntrl->image_buf[((bufno*GA_SPBUFSIZE)>>2) + i] & 0xff);
    }
}

void
pxpmaxCopySpan(ysrc, ydst, xOrigin, xStart, xEnd, alu, planemask)
     int ysrc;
     int ydst;
     int xOrigin;
     int xStart;
     int xEnd;
     int alu;
     int planemask;
{
  int *nbuf_addr;
  int nbuf;
  int i;
  int image_off, span_len;

  span_len = xEnd - xStart;
  if(xOrigin <= xStart)
    span_len--;

  if((span_len >= GA_SPBUFSIZE) || (span_len < 0))
    {
      ErrorF("Span too large or negative\n");
      return;
    }
  if((xStart < 0) || (ysrc < 0) || (ydst < 0) || (xEnd > px_info->max_x)
     || (ysrc >= px_info->max_y) || (ydst >= px_info->max_y))
    {
      ErrorF("CopySpan: bad region\n");
      return;
    }
  
  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return;
    }
  
  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  nbuf_addr[0] = 5;
  nbuf_addr[1] = 11;
  nbuf_addr[2] = 16;
  nbuf_addr[3] = STAMP_CMD_COPYSPANS | LW_PERPKT | CLIPRECT;
  planemask &= 0xff;
  nbuf_addr[4] = (1 << 24) | (planemask << 16) | (planemask << 8) | planemask;
  nbuf_addr[5] = 0;
  nbuf_addr[6] = aluxtopx(alu);
  nbuf_addr[6] |= UPD_ENABLE | WE_CLIPRECT | SPAN_MASK;
  if(xStart == xOrigin)
    nbuf_addr[6] |= COPYSPAN_ALIGNED;
  nbuf_addr[7] = 3;
  nbuf_addr[8] = 0x7c00400;
  nbuf_addr[9] = 0x7ffffff;
  nbuf_addr[10] = span_len << 3;
  nbuf_addr[11] = xOrigin << 19 | (ysrc << 3);
  nbuf_addr[12] = xStart << 19 | (ydst << 3);
  nbuf_addr[13] = span_len << 3;

  ga_send_packet();
}

void
pxpmaxSlowCopySpan(ysrc, ydst, xOrigin, xStart, xEnd, alu, planemask)
     int ysrc;
     int ydst;
     int xOrigin;
     int xStart;
     int xEnd;
     int alu;
     int planemask;
{
  int *nbuf_addr;
  int nbuf;
  int i;
  int image_off, span_len;

  span_len = xEnd - xStart;
  if((span_len >= GA_SPBUFSIZE) || (span_len < 0))
    {
      ErrorF("Span too large or negative\n");
      return;
    }
  if((xStart < 0) || (ysrc < 0) || (ydst < 0) || (xEnd > px_info->max_x)
     || (ysrc >= px_info->max_y) || (ydst >= px_info->max_y))
    {
      ErrorF("SlowCopySpan: bad region\n");
      return;
    }
  
  bufno ^= 1;
  
  image_off = bufno*GA_SPBUFSIZE + GA_IMAGE_BUFFER_OFF - 12;

  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return;
    }
  
  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  nbuf_addr[0] = 5;
  nbuf_addr[1] = 8;
  nbuf_addr[2] = 16;
  nbuf_addr[3] = STAMP_CMD_READSPANS;
  nbuf_addr[4] = (1 << 24) | 0xffffff;
  nbuf_addr[5] = 0;
  nbuf_addr[6] = UMET_NOOP | UPD_ENABLE | SPAN_MASK;
  nbuf_addr[7] = 0x40000000 | px_info->rb_phys | image_off;
  nbuf_addr[8] = 2*span_len*sizeof(int);
  nbuf_addr[9] = xOrigin << 19 | (ysrc << 3);
  nbuf_addr[10] = 0x0;
  
  ga_send_packet();

  nbuf = rb_com2d->lastWritten;
  if(nbuf >= GA_QUEUE_PACKETS)
    {
      ErrorF("Too many packets\n");
      return;
    }
  
  nbuf_addr = &(rb_cntrl->IntrBuf[nbuf][0]);
  nbuf_addr[0] = 5;
  nbuf_addr[1] = 8;
  nbuf_addr[2] = 16;
  nbuf_addr[3] = STAMP_CMD_WRITESPANS;
  planemask &= 0xff;
  nbuf_addr[4] = (1 << 24) | (planemask << 16) | (planemask << 8) | planemask;
  nbuf_addr[5] = 0;
  nbuf_addr[6] = aluxtopx(alu);
  nbuf_addr[6] |= UPD_ENABLE | SPAN_MASK;
  nbuf_addr[7] = px_info->rb_phys | image_off;
  nbuf_addr[8] = 2*span_len*sizeof(int);
  nbuf_addr[9] = xStart << 19 | (ydst << 3);
  nbuf_addr[10] = 0x0;
  
  ga_send_packet();
}
